from unittest.mock import call, patch
from uuid import uuid4

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, DataSource, Transcription
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User
from arkindex_common.enums import TranscriptionType
from arkindex_common.ml_tool import MLToolType


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.src = DataSource.objects.get(slug='test', type=MLToolType.Recognizer)
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.private_corpus.corpus_right.create(user=cls.private_read_user)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def test_create_transcription_require_login(self):
        response = self.client.post(reverse('api:transcription-create', kwargs={'pk': self.line.id}), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Authentication credentials were not provided.'
        })

    def test_create_transcription_write_right(self):
        self.client.force_login(self.private_read_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
            format='json',
            data={'type': TranscriptionType.Word.value, 'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_create_transcription_no_element(self, delay_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': uuid4()}),
            format='json',
            data={'type': TranscriptionType.Word.value, 'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_create_manual_transcription(self, delay_mock):
        """
        Checks the view creates a manual transcription and runs ES indexing
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={'type': TranscriptionType.Line.value, 'text': 'A perfect day in a perfect place'}
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': None,
            'text': 'A perfect day in a perfect place',
            'type': 'line',
            'source': None,
            'worker_version_id': None,
        })

        new_ts = Transcription.objects.get(text='A perfect day in a perfect place', type=TranscriptionType.Line)
        self.assertIsNone(new_ts.score)
        self.assertEqual(new_ts.worker_version, None)
        self.assertTrue(self.line.transcriptions.filter(pk=new_ts.id).exists())

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            element_id=str(self.line.id),
            corpus_id=None,
            entity_id=None,
            transcriptions=True,
            elements=True,
            entities=False,
            drop=False,
            user_id=None,
            description=f'Indexation of element {self.line.id}',
        ))

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_create_duplicated_transcription(self, delay_mock):
        """
        Check the view creates a new manual transcriptions with a similar text and element
        """
        self.client.force_login(self.user)
        ts = self.line.transcriptions.create(text='GLOUBIBOULGA', type=TranscriptionType.Line.value)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'type': 'line',
                    'text': ts.text,
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            Transcription.objects.filter(element=self.line, text=ts.text, worker_version__isnull=True).count(),
            2
        )

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_create_transcription_wrong_type(self, delay_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'type': 'AAAAA', 'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ['Value is not of type TranscriptionType']
        })
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_transcription_no_search(self, delay_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'type': TranscriptionType.Line.value, 'text': 'A classy text line'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(delay_mock.called)

    def test_create_transcription_worker_version_non_internal(self):
        """
        An internal user is required to create a transcription with a worker version
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'worker_version': str(self.worker_version.id),
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': [
                'An internal user is required to create a transcription refering to a worker_version'
            ]
        })

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_create_transcription_worker_version(self, delay_mock):
        """
        Creates a transcription with a worker version triggers its indexation on ElasticSearch
        """
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'worker_version': str(self.worker_version.id),
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='NEKUDOTAYIM')
        self.assertEqual(tr.worker_version, self.worker_version)
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': .42,
            'source': None,
            'text': 'NEKUDOTAYIM',
            'type': 'word',
            'worker_version_id': str(self.worker_version.id),
        })

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            element_id=str(self.line.id),
            corpus_id=None,
            entity_id=None,
            transcriptions=True,
            elements=True,
            entities=False,
            drop=False,
            user_id=None,
            description=f'Indexation of element {self.line.id}',
        ))

    def test_manual_transcription_forbidden_type(self):
        """
        Creating a manual transcription with a non allowed type is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'type': TranscriptionType.Word.value, 'text': 'NEKUDOTAYIM'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Only transcriptions of type 'line' are allowed for this element"]
        })

    def test_manual_transcription_no_allowed_type(self):
        """
        Manual transctiptions are forbidden for corpus types with no allowed transcription
        """
        self.line.type.allowed_transcription = None
        self.line.type.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={'type': TranscriptionType.Line.value, 'text': 'A classy text line'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['The element type does not allow creating a manual transcription']
        })

    def test_ml_trancription_required_score(self):
        """
        A score is required when creating a transcription on an element with an worker version
        """
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'CIRCLE',
                'worker_version': str(self.worker_version.id)
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'score': ['This field is required for transcription with a worker version.']})

    def test_ml_trancription_requires_element_zone(self):
        """
        A worker cannot create a transcription on an element without a zone
        """
        null_zone_page = self.corpus.elements.create(type=self.page.type)
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'type': TranscriptionType.Word.value,
                'text': 'ELLIPSE',
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['The element has no zone']})
