import logging
import urllib.parse

from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.tokens import default_token_generator
from django.core.mail import send_mail
from django.db.models import Count, Q
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.http import urlsafe_base64_encode
from django.views.generic import RedirectView
from django_rq.queues import get_queue
from django_rq.settings import QUEUES
from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed, NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from rest_framework.views import APIView
from rq.job import JobStatus

from arkindex.documents.models import Corpus
from arkindex.project.permissions import IsAuthenticatedOrReadOnly, IsVerified
from arkindex.users.models import Group, LevelRight, OAuthStatus, Scope, User, UserScope
from arkindex.users.providers import get_provider, oauth_providers
from arkindex.users.serializers import (
    EmailLoginSerializer,
    GroupMembershipsSerializer,
    GroupSerializer,
    JobSerializer,
    NewUserSerializer,
    OAuthCredentialsSerializer,
    OAuthProviderClassSerializer,
    OAuthRetrySerializer,
    PasswordResetConfirmSerializer,
    PasswordResetSerializer,
    UpdateUserTranskribusSerializer,
    UserSerializer,
)

logger = logging.getLogger(__name__)


class ProvidersList(ListAPIView):
    """
    List supported OAuth providers
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthProviderClassSerializer
    pagination_class = None
    openapi_overrides = {
        'operationId': 'ListOAuthProviders',
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return list(filter(lambda p: p.enabled(), oauth_providers))


class CredentialsList(ListAPIView):
    """
    List all OAuth credentials for the authenticated user.
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthCredentialsSerializer
    openapi_overrides = {
        'tags': ['oauth'],
    }

    def get_queryset(self):
        qs = self.request.user.credentials.order_by('id')
        if 'status' in self.request.query_params:
            qs = qs.filter(status=self.request.query_params['status'])
        return qs


class CredentialsRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve OAuth credentials.
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthCredentialsSerializer
    openapi_overrides = {
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return self.request.user.credentials.order_by('id')

    def perform_destroy(self, instance):
        instance.provider_class(request=self.request, credentials=instance).disconnect()
        super().perform_destroy(instance)


class UserRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve information about the authenticated user
    """
    permission_classes = (IsAuthenticatedOrReadOnly, )
    serializer_class = UserSerializer
    openapi_overrides = {
        'tags': ['users'],
    }

    def get_object(self):
        return self.request.user

    def retrieve(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            # Include features on GET to let the frontend still know the instance features when not logged in
            return Response(
                {
                    'detail': PermissionDenied.default_detail,
                    'features': settings.ARKINDEX_FEATURES,
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        return super().retrieve(*args, **kwargs)

    def perform_destroy(self, instance):
        logout(self.request)


class UserCreate(CreateAPIView):
    serializer_class = NewUserSerializer
    openapi_overrides = {
        'operationId': 'Register',
        'description': 'Register as a new user.',
        'security': [],
        'tags': ['users'],
    }

    def post(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['signup']:
            raise ValidationError(['Registration has been disabled on this instance.'])
        return super().post(*args, **kwargs)

    def perform_create(self, serializer):
        user = serializer.save()
        login(self.request, user)

        activation_url = '{}?{}'.format(
            self.request.build_absolute_uri(reverse('api:user-token')),
            urllib.parse.urlencode({
                'email': user.email,
                'token': default_token_generator.make_token(user),
            }),
        )
        sent = send_mail(
            # Subject cannot have any newlines
            subject=''.join(render_to_string('registration/verification_email_subject.txt').splitlines()),
            message=render_to_string(
                'registration/verification_email.html',
                context={
                    'url': activation_url,
                },
                request=self.request,
            ),
            from_email=None,
            recipient_list=[user.email],
            fail_silently=True,
        )
        if sent == 0:
            logger.error('Failed to send registration email to {}'.format(user.email))


class UserEmailLogin(CreateAPIView):
    """
    Login using an email and a password
    """
    serializer_class = EmailLoginSerializer
    openapi_overrides = {
        'operationId': 'Login',
        'security': [],
        'tags': ['users'],
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            user = User.objects.get_by_natural_key(serializer.validated_data['email'])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not user.has_usable_password() or not user.check_password(serializer.validated_data['password']):
            raise AuthenticationFailed

        login(self.request, user)

        headers = self.get_success_headers(serializer.data)
        return Response(
            UserSerializer(user).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


class UserEmailVerification(APIView):
    """
    Verify a user's email address
    """

    openapi_overrides = {
        'operationId': 'VerifyEmail',
        'tags': ['users'],
        'parameters': [
            {
                'name': 'email',
                'in': 'query',
                'description': 'E-mail to verify',
                'required': True,
                'schema': {
                    'type': 'string',
                    'format': 'email',
                },
            },
            {
                'name': 'token',
                'in': 'query',
                'description': 'Verification token',
                'required': True,
                'schema': {
                    'type': 'string',
                },
            },
        ],
    }

    def get(self, *args, **kwargs):
        if not all(arg in self.request.GET for arg in ('email', 'token')):
            raise ValidationError

        try:
            user = User.objects.get_by_natural_key(self.request.GET['email'])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not default_token_generator.check_token(user, self.request.GET['token']):
            raise AuthenticationFailed

        user.verified_email = True
        user.save()
        UserScope.objects.bulk_create([
            UserScope(user=user, scope=Scope.CreateIIIFImage),
            UserScope(user=user, scope=Scope.UploadS3Image),
        ])

        if user.has_usable_password():
            login(self.request, user)
            # A user with a verified email is granted a private trial corpus with admin rights
            if not Corpus.objects.admin(user=user).exists():
                corpus = Corpus.objects.create(
                    name='My Project',
                    description=f'Project for {user.email}'
                )
                corpus.corpus_right.create(
                    user=user,
                    can_admin=True,
                    can_write=True
                )
                # Add default types
                corpus.create_default_types()

            # Redirtect to home page
            url = '/'
        else:
            url = reverse('password_reset_confirm', kwargs={
                'uidb64': urlsafe_base64_encode(str(user.id).encode()),
                'token': self.request.GET['token'],
            })

        return Response(
            status=status.HTTP_302_FOUND,
            headers={'Location': url},
        )


class PasswordReset(CreateAPIView):
    """
    Start a password reset flow
    """
    serializer_class = PasswordResetSerializer
    openapi_overrides = {
        'operationId': 'ResetPassword',
        'security': [],
        'tags': ['users'],
    }

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        form = PasswordResetForm(serializer.validated_data)

        if form.is_valid():
            form.save(
                request=self.request,
                use_https=self.request.is_secure(),
                token_generator=default_token_generator,
                email_template_name='registration/password_reset_email.html',
                subject_template_name='registration/password_reset_subject.txt',
            )
            headers = self.get_success_headers(serializer.data)
            return Response(
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        else:
            return Response(
                form.errors.get_json_data(),
                status=status.HTTP_400_BAD_REQUEST,
                headers=self.default_response_headers
            )


class PasswordResetConfirm(CreateAPIView):
    """
    Confirm a password reset using data from the confirmation e-mail
    """
    serializer_class = PasswordResetConfirmSerializer
    openapi_overrides = {
        'operationId': 'PasswordResetConfirm',
        'security': [],
        'tags': ['users'],
    }


class UpdateUserTranskribus(UpdateAPIView):
    """
    Update and validate Transkribus account
    Only the user's email is stored in our database
    """
    permission_classes = (IsVerified, )
    serializer_class = UpdateUserTranskribusSerializer
    openapi_overrides = {
        'tags': ['users'],
    }

    def get_object(self):
        return self.request.user


class OAuthSignIn(APIView):
    """
    Start the OAuth authentication code flow for a given provider
    """
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'StartOAuthSignIn',
        'tags': ['oauth'],
    }

    def get(self, *args, **kwargs):
        if 'provider' not in kwargs:
            raise ValidationError('Missing provider')
        provider_class = get_provider(kwargs['provider'])
        if not provider_class:
            raise ValidationError('Unknown provider')
        url = self.request.GET.get('url', provider_class.url)
        if url:
            parsed = urllib.parse.urlparse(url)
            if (parsed.scheme != 'https'
                    or parsed.netloc == ''
                    or parsed.params != ''
                    or parsed.query != ''
                    or parsed.fragment != ''):
                raise ValidationError('Invalid GitLab instance URL')
        # Create OAuthCredentials without a token
        creds = self.request.user.credentials.create(
            provider_name=kwargs['provider'],
            provider_url=url,
        )
        return Response({
            'url': provider_class(request=self.request, credentials=creds).get_authorize_uri(),
        })


class OAuthRetry(RetrieveAPIView):
    """
    Retry the OAuth authentication code flow for pending credentials
    """
    permission_classes = (IsVerified, )
    serializer_class = OAuthRetrySerializer
    openapi_overrides = {
        'operationId': 'RetryOAuthCredentials',
        'tags': ['oauth'],
    }

    def get_queryset(self):
        return self.request.user.credentials

    def get_object(self):
        creds = super().get_object()
        if creds.status == OAuthStatus.Done:
            self.permission_denied()
        return creds

    def retrieve(self, request, *args, **kwargs):
        creds = self.get_object()
        return Response({
            'url': creds.provider_class(request=self.request, credentials=creds).get_authorize_uri(),
        })


class OAuthCallback(UserPassesTestMixin, RedirectView):
    """
    Callback for OAuth responses
    """
    pattern_name = 'credentials'
    permission_classes = (IsVerified, )
    raise_exception = True

    def test_func(self):
        """
        Perform Django REST framework permission checks, but without REST framework.
        """
        return all(perm().has_permission(self.request, self) for perm in self.permission_classes)

    def get(self, request, *args, **kwargs):
        assert 'provider' in kwargs
        provider_class = get_provider(kwargs['provider'])
        if not provider_class:
            raise ValueError('Unknown provider')
        provider = provider_class(self.request)
        try:
            provider.handle_callback()
            provider.credentials.status = OAuthStatus.Done
        except Exception as e:
            if not provider.credentials:
                raise
            provider.credentials.status = OAuthStatus.Error
            logger.warning(f'OAuth callback error: {e}', exc_info=e)
        provider.credentials.save()
        return super().get(request)


class GroupsList(ListCreateAPIView):
    """
    List existing groups either if they are public or the request user is a member.
    A POST request allow to create a new group.
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'security': [],
        'tags': ['users'],
    }

    def get_queryset(self):
        # List public groups and ones to which user belongs
        return Group.objects \
            .annotate(members_count=Count('users')) \
            .filter(Q(public=True) | Q(users__in=[self.request.user])) \
            .order_by('name', 'id')


class JobList(ListAPIView):
    """
    List asynchronous jobs linked to the current user.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer
    pagination_class = None

    def get_queryset(self):
        if not self.request:
            return []
        return self.request.user.get_rq_jobs()


class JobRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve a single job by ID.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer

    def get_object(self):
        for queue_name in QUEUES.keys():
            job = get_queue(queue_name).fetch_job(str(self.kwargs['pk']))
            if not job:
                continue
            if job.user_id != self.request.user.id:
                raise NotFound
            return job
        raise NotFound

    def perform_destroy(self, instance):
        if instance.get_status(refresh=False) == JobStatus.STARTED:
            raise ValidationError(['Cannot delete a running job.'])
        instance.delete()


class GroupMembershipsList(ListAPIView):
    """
    List members of a group to which user belongs with their privileges
    """
    serializer_class = GroupMembershipsSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'security': [],
        'tags': ['users'],
    }

    def get_queryset(self):
        group = get_object_or_404(Group.objects.filter(
            # Filter groups to which user belongs
            users__in=[self.request.user],
            id=self.kwargs['pk']
        ))
        return group.memberships.order_by('user__display_name', 'id').prefetch_related('user')


class GroupDetails(RetrieveUpdateDestroyAPIView):
    """
    Retrieve details about a specific group
    """
    serializer_class = GroupSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'security': [],
        'tags': ['users'],
    }

    def get_object(self):
        if not hasattr(self, '_group'):
            self._group = super().get_object()
        return self._group

    def check_object_permissions(self, request, obj):
        if not hasattr(self, '_member'):
            self._member = obj.memberships.get(user=self.request.user)
        # Check the user has the right to delete or update a group
        super().check_object_permissions(request, obj)
        user_level = LevelRight.from_level(self._member.level)
        if self.request.method not in SAFE_METHODS and user_level is not LevelRight.Admin:
            raise self.permission_denied(
                request,
                message='Only members with an admin privilege may update or delete this group.'
            )

    def get_queryset(self):
        # Filter groups to which user belongs only
        return Group.objects\
            .annotate(members_count=Count('memberships')) \
            .filter(Q(users__in=[self.request.user]))
